"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: bart@inmanta.com
"""

import os
import pwd
import shutil
import subprocess
import tempfile

import pytest

import grp
from inmanta.agent.io import parse_agent_uri
from inmanta.agent.io.local import BashIO
from inmanta.agent.io.local import LocalIO


io_list = [LocalIO("local:", {}), BashIO("local:", {}), BashIO("local:", {}, run_as="root")]
io_names = ["local", "bash", "bash_root"]


@pytest.yield_fixture
def testdir():
    testdir = tempfile.mkdtemp()
    yield testdir
    shutil.rmtree(testdir)


@pytest.mark.parametrize("io", io_list)
def test_check_hash(io, testdir):
    filename = os.path.join(testdir, "hashfile")
    with open(filename, "w+") as fd:
        fd.write("test")

    assert io.hash_file(filename) == "a94a8fe5ccb19ba61c4c0873d391e987982fbbd3"


@pytest.mark.parametrize("io", io_list)
def test_check_read(io, testdir):
    test_str = "hello é"
    filename = os.path.join(testdir, "readfile")
    with open(filename, "wb+") as fd:
        fd.write(test_str.encode())

    assert io.read(filename) == test_str


@pytest.mark.parametrize("io", io_list)
def test_check_read_binary(io, testdir):
    test_str = b'\1\2\3\4\5\6'
    filename = os.path.join(testdir, "readbinaryfile")
    with open(filename, "wb+") as fd:
        fd.write(test_str)

    result = io.read_binary(filename)

    assert isinstance(result, bytes)
    assert result == test_str


@pytest.mark.parametrize("io", io_list, ids=io_names)
def test_check_run_pipe_actual(io, testdir):
    filename = os.path.join(testdir, "testfile")
    result = io.run('sh', ['-c', 'export PATH=$PATH:/usr/lib/jvm/jre-1.7.0-openjdk/bin; env >' + filename + ' 2>&1'])
    print(result)
    assert "/usr/lib/jvm/jre-1.7.0-openjdk/bin" not in result[0]
    assert result[2] == 0
    assert os.path.exists(filename)


@pytest.mark.parametrize("io", io_list)
def test_check_run(io, testdir):
    result = io.run("echo", ["world"])
    assert result[2] == 0
    assert result[0] == "world"

    # check cwd
    os.makedirs(os.path.join(testdir, "check_run" + str(io)))
    out, _, _ = io.run("ls", [], cwd=testdir)
    assert "check_run" + str(io) in out


@pytest.mark.parametrize("io", io_list)
def test_check_run_env(io, testdir):
    value = "testenvmagicstring"
    result = io.run("env", env={"TEST_ENV": value})

    assert value in result[0]


@pytest.mark.parametrize("io", io_list)
def test_check_file_exists(io, testdir):
    assert io.file_exists(testdir)
    assert not io.file_exists("/blablablablablalbla")

    filename = os.path.join(testdir, "testfile" + str(io))
    assert not io.file_exists(filename)

    with open(filename, "w+") as fd:
        fd.write("")

    assert io.file_exists(filename)


@pytest.mark.parametrize("io", io_list)
def test_check_symlink(io, testdir):
    dst = os.path.join(testdir, "ln" + str(io))
    io.symlink(testdir, dst)

    assert os.path.exists(dst)
    assert os.path.islink(dst)
    assert os.readlink(dst) == testdir


@pytest.mark.parametrize("io", io_list)
def test_check_readlink(io, testdir):
    dst = os.path.join(testdir, "rdln" + str(io))
    os.symlink(testdir, dst)

    assert io.readlink(dst) == testdir


@pytest.mark.parametrize("io", io_list)
def test_check_is_symlink(io, testdir):
    dst = os.path.join(testdir, "isln" + str(io))
    os.symlink(testdir, dst)

    assert io.is_symlink(dst)
    assert not io.is_symlink(testdir)


@pytest.mark.parametrize("io", io_list)
def test_check_mkdir(io, testdir):
    mkdir = os.path.join(testdir, "dir" + str(io))
    io.mkdir(mkdir)

    assert os.path.exists(mkdir)
    assert os.path.isdir(mkdir)


@pytest.mark.parametrize("io", io_list)
def test_check_rmdir(io, testdir):
    path = os.path.join(testdir, "rmdir" + str(io))
    os.mkdir(path)
    assert(os.path.exists(path))

    path2 = os.path.join(path, "testfile" + str(io))
    with open(path2, "w+") as fd:
        fd.write("")
    assert(os.path.exists(path2))

    io.rmdir(path)

    assert not os.path.exists(path)
    assert not os.path.exists(path2)


@pytest.mark.parametrize("io", io_list)
def test_check_filestat(io, testdir):
    stat = io.file_stat("/etc")
    assert stat["permissions"] == 755
    assert stat["owner"] == "root"
    assert stat["group"] == "root"


@pytest.mark.parametrize("io", io_list)
def test_check_chown(io, testdir):
    # chown to the same user so we do not need root to run this test
    path = os.path.join(testdir, "chown" + str(io))
    with open(path, "w+") as fd:
        fd.write("")

    assert os.path.exists(path)

    user = pwd.getpwuid(os.getuid())[0]
    groups = [grp.getgrgid(g).gr_name for g in os.getgroups()]

    if len(groups) == 0:
        # fix for jenkins docker
        groups = [user]

    io.chown(path, user)
    stat = io.file_stat(path)
    assert stat["owner"] == user

    io.chown(path, user, groups[0])
    stat = io.file_stat(path)
    assert stat["owner"] == user
    assert stat["group"] == groups[0]

    io.chown(path, None, groups[0])
    stat = io.file_stat(path)
    assert stat["group"] == groups[0]


@pytest.mark.parametrize("io", io_list)
def test_check_put(io, testdir):
    path = os.path.join(testdir, "put" + str(io))
    io.put(path, b"hello")

    with open(path, "rb") as fd:
        assert b"hello" == fd.read()


@pytest.mark.parametrize("io", io_list)
def test_check_chmod(io, testdir):
    path = os.path.join(testdir, "chmod" + str(io))
    with open(path, "w+") as fd:
        fd.write("Test\n")

    st = os.stat(path)
    mode = str(oct(st.st_mode))[-4:]

    io.chmod(path, "0755")

    st = os.stat(path)
    new_mode = str(oct(st.st_mode))[-4:]

    assert mode != new_mode
    assert "0755" == new_mode


@pytest.mark.parametrize("io", io_list)
def test_hash_dir(io, testdir):
    dir_path = os.path.join(testdir, "dir")
    if not os.path.exists(dir_path):
        io.mkdir(dir_path)

    with pytest.raises(Exception):
        io.hash_file(dir_path)


@pytest.mark.parametrize("io", io_list)
def test_timeout(io):
    with pytest.raises(subprocess.TimeoutExpired):
        io.run("sleep", ["0.1"], timeout=0.05)


def test_uri_parse():
    scheme, config = parse_agent_uri("local:")
    assert scheme == "local"
    assert len(config) == 0

    scheme, config = parse_agent_uri("ssh://ec2-user@1.2.3.4:22")
    assert scheme == "ssh"
    assert len(config) == 3
    assert config["user"] == "ec2-user"
    assert config["port"] == "22"
    assert config["host"] == "1.2.3.4"

    scheme, config = parse_agent_uri("")
    assert scheme == "local"
    assert len(config) == 0

    scheme, config = parse_agent_uri("hostname")
    assert scheme == "ssh"
    assert len(config) == 3
    assert config["host"] == "hostname"

    scheme, config = parse_agent_uri("localhost")
    assert scheme == "local"

    scheme, config = parse_agent_uri("ssh://ec2-user@1.2.3.4:22?python=/usr/bin/python2")
    assert scheme == "ssh"
    assert len(config) == 4
    assert config["user"] == "ec2-user"
    assert config["port"] == "22"
    assert config["host"] == "1.2.3.4"
    assert config["python"] == "/usr/bin/python2"



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

# pylint: disable-msg=W0613,R0201

import logging

from inmanta.ast.statements import GeneratorStatement
from inmanta.const import LOG_LEVEL_TRACE
from inmanta.execute.util import Unknown
from inmanta.execute.runtime import ExecutionContext, Resolver, QueueScheduler, ResultVariable, ResultCollector
from inmanta.ast import RuntimeException, TypingException, NotFoundException, Location, Namespace, DuplicateException,\
    LocatableString, TypeReferenceAnchor, AttributeReferenceAnchor
from inmanta.execute.tracking import ImplementsTracker
from typing import List, Dict, Tuple, Set  # noqa: F401
from inmanta.ast.statements import ExpressionStatement
from inmanta.ast.blocks import BasicBlock
from inmanta.ast.statements.assign import SetAttributeHelper

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from inmanta.ast.entity import Default, Entity, Implement, EntityLike  # noqa: F401

LOGGER = logging.getLogger(__name__)


class SubConstructor(GeneratorStatement):
    """
        This statement selects an implementation for a given object and
        imports the statements
    """

    def __init__(self, instance_type: "Entity", implements: "Implement") -> None:
        GeneratorStatement.__init__(self)
        self.type = instance_type
        self.location = instance_type.get_location()
        self.implements = implements

    def normalize(self) -> None:
        # done in define type
        pass

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, ResultVariable]:
        try:
            resv = resolver.for_namespace(self.implements.constraint.namespace)
            return self.implements.constraint.requires_emit(resv, queue)
        except NotFoundException as e:
            e.set_statement(self.implements)
            raise e

    def execute(self, requires: Dict[object, object], instance: Resolver, queue: QueueScheduler) -> object:
        """
            Evaluate this statement
        """
        LOGGER.log(LOG_LEVEL_TRACE, "executing subconstructor for %s implement %s", self.type, self.implements.location)
        expr = self.implements.constraint
        if not expr.execute(requires, instance, queue):
            return None

        myqueue = queue.for_tracker(ImplementsTracker(self, instance))

        implementations = self.implements.implementations

        for impl in implementations:
            if instance.add_implementation(impl):
                # generate a subscope/namespace for each loop
                xc = ExecutionContext(impl.statements, instance.for_namespace(impl.statements.namespace))
                xc.emit(myqueue)

        return None

    def __repr__(self) -> str:
        return "SubConstructor(%s)" % self.type


class GradualFor(ResultCollector):
    # this class might be unnecessary if receive-result is always called and exactly once

    def __init__(self, stmt: "For", resolver: Resolver, queue: QueueScheduler):
        self.resolver = resolver
        self.queue = queue
        self.stmt = stmt
        self.seen = set()  # type: Set[int]

    def receive_result(self, value, location):
        if id(value) in self.seen:
            return
        self.seen.add(id(value))

        xc = ExecutionContext(self.stmt.module, self.resolver.for_namespace(self.stmt.module.namespace))
        loopvar = xc.lookup(self.stmt.loop_var)
        loopvar.set_provider(self.stmt)
        loopvar.set_value(value, self.stmt.location)
        xc.emit(self.queue)


class For(GeneratorStatement):
    """
        A for loop
    """

    def __init__(self, variable: ExpressionStatement, loop_var: LocatableString, module: BasicBlock) -> None:
        GeneratorStatement.__init__(self)
        self.base = variable
        self.loop_var = str(loop_var)
        self.loop_var_loc = loop_var.get_location()
        self.module = module
        self.anchors.extend(module.get_anchors())
        self.anchors.extend(variable.get_anchors())

    def __repr__(self) -> str:
        return "For(%s)" % self.loop_var

    def normalize(self) -> None:
        self.base.normalize()
        # self.loop_var.normalize(resolver)
        self.module.normalize()
        self.module.add_var(self.loop_var)

    def requires(self) -> List[str]:
        base = self.base.requires()
        var = self.loop_var
        ext = self.module.requires
        return list(set(base).union(ext) - set(var))

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, ResultVariable]:
        # pass context via requires!
        helper = GradualFor(self, resolver, queue)

        helperwrapped = ResultVariable()
        helperwrapped.set_value(helper, self.location)

        basereq = self.base.requires_emit_gradual(resolver, queue, helper)
        basereq[self] = helperwrapped

        return basereq

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        """
            Evaluate this statement.
        """
        var = self.base.execute(requires, resolver, queue)

        if isinstance(var, Unknown):
            return None

        if not isinstance(var, list):
            raise TypingException(self, "A for loop can only be applied to lists and relations")

        helper = requires[self]

        for loop_var in var:
            # generate a subscope/namespace for each loop
            helper.receive_result(loop_var, self.location)

        return None


class Constructor(GeneratorStatement):
    """
        This class represents the usage of a constructor to create a new object.

        @param class_type: The type of the object that is created by this
            constructor call.
    """

    def __init__(self,
                 class_type: LocatableString,
                 attributes: List[Tuple[LocatableString, ExpressionStatement]],
                 location: Location,
                 namespace: Namespace) -> None:
        GeneratorStatement.__init__(self)
        self.class_type = str(class_type)
        self.__attributes = {}  # type: Dict[str,ExpressionStatement]
        self.implemented = False
        self.register = False
        self.location = location
        self.namespace = namespace
        self.anchors.append(TypeReferenceAnchor(class_type.get_location(), namespace, str(class_type)))
        for a in attributes:
            self.add_attribute(a[0], a[1])

        self._direct_attributes = {}  # type: Dict[str,ExpressionStatement]
        self._indirect_attributes = {}  # type: Dict[str,ExpressionStatement]

    def pretty_print(self) -> str:
        return "%s(%s)" % (self.class_type, ",".join(("%s=%s" % (k, v.pretty_print()) for k, v in self.attributes.items())))

    def normalize(self) -> None:
        mytype = self.namespace.get_type(self.class_type)

        self.type = mytype  # type: EntityLike

        for (k, v) in self.__attributes.items():
            v.normalize()

        inindex = set()

        all_attributes = dict(self.type.get_default_values())
        all_attributes.update(self.__attributes)

        # now check that all variables that have indexes on them, are already
        # defined and add the instance to the index
        for index in self.type.get_entity().get_indices():
            for attr in index:
                if attr not in all_attributes:
                    raise TypingException(self, "%s is part of an index and should be set in the constructor." % attr)
                inindex.add(attr)

        for (k, v) in all_attributes.items():
            attribute = self.type.get_entity().get_attribute(k)
            if attribute is None:
                raise TypingException(self, "no attribute %s on type %s" % (k, self.type.get_full_name()))
            if k not in inindex:
                self._indirect_attributes[k] = v
            else:
                self._direct_attributes[k] = v

    def requires(self) -> List[str]:
        out = [req for (k, v) in self.__attributes.items() for req in v.requires()]
        out.extend([req for (k, v) in self.get_default_values().items() for req in v.requires()])
        return out

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, ResultVariable]:
        # direct
        direct = [x for x in self._direct_attributes.items()]

        direct_requires = {rk: rv for (k, v) in direct for (rk, rv) in v.requires_emit(resolver, queue).items()}
        LOGGER.log(LOG_LEVEL_TRACE,
                   "emitting constructor for %s at %s with %s",
                   self.class_type,
                   self.location,
                   direct_requires)

        return direct_requires

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler):
        """
            Evaluate this statement.
        """
        LOGGER.log(LOG_LEVEL_TRACE, "executing constructor for %s at %s", self.class_type, self.location)

        # the type to construct
        type_class = self.type.get_entity()

        # the attributes
        attributes = {k: v.execute(requires, resolver, queue) for (k, v) in self._direct_attributes.items()}

        # check if the instance already exists in the index (if there is one)
        instances = []
        for index in type_class.get_indices():
            params = []
            for attr in index:
                params.append((attr, attributes[attr]))

            obj = type_class.lookup_index(params, self)

            if obj is not None:
                if obj.get_type().get_entity() != type_class:
                    raise DuplicateException(self, object, "Type found in index is not an exact match")
                instances.append(obj)

        if len(instances) > 0:
            # ensure that instances are all the same objects
            first = instances[0]
            for i in instances[1:]:
                if i != first:
                    raise Exception("Inconsistent indexes detected!")

            object_instance = first
            self.copy_location(object_instance)
            for k, v in attributes.items():
                object_instance.set_attribute(k, v, self.location)
        else:
            # create the instance
            object_instance = type_class.get_instance(attributes, resolver, queue, self.location)

        # deferred execution for indirect attributes
        for attributename, valueexpression in self._indirect_attributes.items():
            var = object_instance.get_attribute(attributename)
            reqs = valueexpression.requires_emit_gradual(resolver, queue, var)
            SetAttributeHelper(queue, resolver, var, reqs, valueexpression, self, object_instance, attributename)

        # add anonymous implementations
        if self.implemented:
            # generate an import for the module
            raise Exception("don't know this feature")

        else:
            # generate an implementation
            for stmt in type_class.get_sub_constructor():
                stmt.emit(object_instance, queue)

        if self.register:
            raise Exception("don't know this feature")

        object_instance.trackers.append(queue.get_tracker())

        return object_instance

    def add_attribute(self, lname: LocatableString, value: ExpressionStatement) -> None:
        """
            Add an attribute to this constructor call
        """
        name = str(lname)
        if name not in self.__attributes:
            self.__attributes[name] = value
            self.anchors.append(AttributeReferenceAnchor(lname.get_location(), lname.namespace, self.class_type, name))
            self.anchors.extend(value.get_anchors())
        else:
            raise RuntimeException(self, "The attribute %s in the constructor call of %s is already set."
                                   % (name, self.class_type))

    def get_attributes(self) -> Dict[str, ExpressionStatement]:
        """
            Get the attribtues that are set for this constructor call
        """
        return self.__attributes

    attributes = property(get_attributes)

    def __repr__(self) -> str:
        """
            The representation of the this statement
        """
        return "Construct(%s)" % (self.class_type)



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from inmanta import agent
import pytest
from utils import retry_limited
from inmanta.agent import reporting
from inmanta.server import SLICE_SESSION_MANAGER


@pytest.mark.slowtest
@pytest.mark.asyncio
async def test_agent_get_status(server, environment):
    myagent = agent.Agent(hostname="node1", environment=environment, agent_map={"agent1": "localhost"}, code_loader=False)
    myagent.add_end_point_name("agent1")
    await myagent.start()

    await retry_limited(lambda: len(server.get_slice(SLICE_SESSION_MANAGER)._sessions) == 1, 0.5)
    clients = server.get_slice(SLICE_SESSION_MANAGER)._sessions.values()
    assert len(clients) == 1
    clients = [x for x in clients]
    client = clients[0].get_client()
    status = await client.get_status()
    status = status.get_result()
    for name in reporting.reports.keys():
        assert name in status and status[name] != "ERROR"



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from inmanta.ast import Namespace, TypeNotFoundException, RuntimeException, Locatable, Named, DuplicateException, Location
from inmanta.execute.util import AnyType, NoneValue
import numbers
from typing import Optional


class BasicResolver(object):

    def __init__(self, types):
        self.types = types

    def get_type(self, namespace, name):
        if not isinstance(name, str):
            raise Exception("Should Not Occur, bad AST construction")
        if "::" in name:
            if name in self.types:
                return self.types[name]
            else:
                raise TypeNotFoundException(name, namespace)
        elif name in TYPES:
            return self.types[name]
        else:
            cns = namespace
            while cns is not None:
                full_name = "%s::%s" % (cns.get_full_name(), name)
                if full_name in self.types:
                    return self.types[full_name]
                cns = cns.get_parent()
                raise TypeNotFoundException(name, namespace)


class NameSpacedResolver(object):

    def __init__(self, ns):
        self.ns = ns

    def get_type(self, name):
        return self.ns.get_type(name)

    def get_resolver_for(self, namespace: Namespace):
        return NameSpacedResolver(namespace)


class CastException(Exception):
    """
        This exception is thrown when a type is unable to cast a value to its
        representation.
    """


class Type(Locatable):
    """
        This class is the base class for all types that represent basic data.
        These are types that are not relations.

        This code is rather tricky, as not all types are used in the same way

        1. types that are used as python types and inmanta types
            (i.e. an inmanta instance === python instance of the python class === the inmanta type)
            1. List
            2. Dict
        2. types for which the inmanta type == the python class and
            inmanta instances are python instances of a the corresponding python type
            1. Bool  => bool
            2. Number => float/int
            3. String => str
        3. types for which the inmanta type is an instance of the python type and
            inmanta instance is also an instance of the python type
            1. TypedList
        4. types for which the inmanta type is an instance of the python type and
            inmanta instance is an instance of a different python type
            1. ConstraintType
            2. NullableType

    """

    @classmethod
    def validate(cls, value: object) -> None:
        """
            Validate the given value to check if it satisfies the constraints
            associated with this type
        """
        raise NotImplementedError()

    @classmethod
    def cast(cls, value):
        """
            Cast the given value to this type. If this fails a CastException
            is thrown.

            @param value: The value to cast
        """
        raise NotImplementedError()

    def type_string(self):
        """get the name of the type """
        raise NotImplementedError()

    def __str__(self):
        """get the string representation of the instance of the type """
        raise NotImplementedError(type(self))

    def normalize(self):
        pass

    def is_primitive(self):
        return False


class NamedType(Type, Named):

    def get_double_defined_exception(self, other: "NamedType") -> DuplicateException:
        """produce a customized error message for this type"""
        raise NotImplementedError()


class NullableType(Type):

    def __init__(self, basetype):
        Type.__init__(self)
        self.basetype = basetype

    def cast(self, value):
        """
            Cast the value to the basetype of this constraint
        """
        return self.basetype.cast(value)

    def validate(self, value):
        """
            Validate the given value to check if it satisfies the constraint and
            the basetype.
        """
        if isinstance(value, NoneValue):
            return True

        return self.basetype.validate(value)

    def type_string(self):
        return "%s?" % (self.basetype.type_string())

    def __str__(self):
        return "%s" % (self.basetype)


class Number(Type):
    """
        This class represents an integer or float in the configuration model. On
        these integers the following operations are supported:

        +, -, /, *
    """

    def __init__(self):
        Type.__init__(self)

    @classmethod
    def validate(cls, value):
        """
            Validate the given value to check if it satisfies the constraints
            associated with this type
        """
        if isinstance(value, AnyType):
            return True

        if not isinstance(value, numbers.Number):
            raise RuntimeException(None, "Invalid value '%s', expected Number" % value)

        return True  # allow this function to be called from a lambda function

    @classmethod
    def cast(cls, value):
        """
            Cast the value to a number.

            :see CastableType#cast
        """
        if value is None:
            return value

        try:
            fl_value = float(value)
            try:
                int_value = int(value)
            except ValueError:
                int_value = 0

            if fl_value == int_value:
                return int_value

            return fl_value
        except ValueError:
            raise CastException()

    @classmethod
    def __str__(cls):
        return "number"

    @classmethod
    def is_primitive(cls):
        return True

    @classmethod
    def get_location(cls) -> Location:
        return None

    @classmethod
    def type_string(cls):
        return "number"


class Bool(Type):
    """
        This class represents a simple boolean that can hold true or false.
    """

    def __init__(self):
        Type.__init__(self)

    @classmethod
    def validate(cls, value):
        """
            Validate the given value to check if it satisfies the constraints
            associated with this type
        """
        if isinstance(value, AnyType):
            return True
        if isinstance(value, bool):
            return True
        else:
            raise RuntimeException(None, "Invalid value '%s', expected Bool" % value)

    @classmethod
    def cast(cls, value):
        """
            Convert the given value to value that can be used by the operators
            defined on this type.
        """
        if (value == "true" or value == "True" or value == 1 or value == "1" or value is True):
            return True

        if (value == "false" or value == "False" or value == 0 or value == "0" or value is False):
            return False

        raise CastException()

    @classmethod
    def type_string(cls):
        return "bool"

    @classmethod
    def is_primitive(cls):
        return True

    @classmethod
    def get_location(cls) -> Location:
        return None


class String(Type, str):
    """
        This class represents a string type in the configuration model.
    """

    def __init__(self):
        Type.__init__(self)
        str.__init__(self)

    @classmethod
    def cast(cls, value):
        """
            Cast the given value to a string

            :see CastableType#cast
        """
        return str(value)

    @classmethod
    def validate(cls, value):
        """
            Validate the given value to check if it satisfies the constraints
            associated with this type
        """
        if isinstance(value, AnyType):
            return True
        if not isinstance(value, str):
            raise RuntimeException(None, "Invalid value '%s', expected String" % value)

        return True

    @classmethod
    def type_string(cls):
        return "string"

    @classmethod
    def is_primitive(cls):
        return True

    @classmethod
    def get_location(cls) -> Location:
        return None


class TypedList(Type):

    def __init__(self, basetype):
        Type.__init__(self)
        self.basetype = basetype

    def cast(self, value):
        """
            Cast the value to the basetype of this constraint
        """
        return list([self.basetype.cast(x) for x in value])

    def validate(self, value):
        """
            Validate the given value to check if it satisfies the constraint and
            the basetype.
        """
        if isinstance(value, AnyType):
            return True

        if value is None:
            return True

        if not isinstance(value, list):
            raise RuntimeException(None, "Invalid value '%s', expected list" % value)

        for x in value:
            self.basetype.validate(x)

        return True

    def type_string(self):
        return "%s[]" % (self.type_string())

    def __str__(self):
        return str(self.basetype)


class List(Type, list):
    """
        This class represents a list type in the configuration model. (instances represent instances)
    """

    def __init__(self):
        Type.__init__(self)
        list.__init__(self)

    @classmethod
    def cast(cls, value):
        """
            Cast the given value to a string

            :see CastableType#cast
        """
        return list(value)

    @classmethod
    def validate(cls, value):
        """
            Validate the given value to check if it satisfies the constraints
            associated with this type
        """
        if value is None:
            return True

        if isinstance(value, AnyType):
            return True

        if not isinstance(value, list):
            raise RuntimeException(None, "Invalid value '%s', expected list" % value)

        return True

#     @classmethod
#     def __str__(cls):
#         return "list"

    @classmethod
    def type_string(cls):
        return "list"

    def __str__(self):
        return list.__str__(self)

    @classmethod
    def get_location(cls) -> Location:
        return None


class Dict(Type, dict):
    """
        This class represents a list type in the configuration model. (instances represent instances)
    """

    def __init__(self):
        Type.__init__(self)
        dict.__init__(self)

    @classmethod
    def cast(cls, value):
        """
            Cast the given value to a string

            :see CastableType#cast
        """
        return dict(value)

    @classmethod
    def validate(cls, value):
        """
            Validate the given value to check if it satisfies the constraints
            associated with this type
        """
        if isinstance(value, AnyType):
            return True

        if value is None:
            return True

        if not isinstance(value, dict):
            raise RuntimeException(None, "Invalid value '%s', expected dict" % value)

        return True

    @classmethod
    def type_string(cls):
        return "dict"

    def __str__(self):
        return dict.__str__(self)

    @classmethod
    def get_location(cls) -> Location:
        return None


class ConstraintType(NamedType):
    """
        A type that is based on Number or String but defines additional constraint on this type.
        These constraints only apply on the value of the type.

        The constraint on this type is defined by a regular expression.
    """
    comment: Optional[str]

    def __init__(self, namespace, name):
        Type.__init__(self)

        self.basetype = None  # : ConstrainableType
        self._constraint = None
        self.name = name
        self.namespace = namespace
        self.comment = None

    def set_constraint(self, expression):
        """
            Set the constraint for this type. This baseclass for constraint
            types requires the constraint to be set as a regex that can be
            compiled.
        """
        self.expression = expression
        self._constraint = create_function(expression)

    def get_constaint(self):
        """
            Get the string representation of the constraint
        """
        return self._constraint

    constraint = property(get_constaint, set_constraint)

    def cast(self, value):
        """
            Cast the value to the basetype of this constraint
        """
        self.__base_type.cast(value)

    def validate(self, value):
        """
            Validate the given value to check if it satisfies the constraint and
            the basetype.
        """
        if isinstance(value, AnyType):
            return True

        self.basetype.validate(value)

        if not self._constraint(value):
            raise RuntimeException(self, "Invalid value '%s', constraint does not match" % value)

        return True

    def type_string(self):
        return "%s::%s" % (self.namespace, self.name)

    def __str__(self):
        return self.type_string()

    def get_full_name(self) -> str:
        return self.namespace.get_full_name() + "::" + self.name

    def get_namespace(self) -> "Namespace":
        return self.namespace

    def get_double_defined_exception(self, other: "NamedType") -> DuplicateException:
        return DuplicateException(
            self, other, "TypeConstraint %s is already defined" % (self.get_full_name()))


def create_function(expression):
    """
        Function that returns a function that evaluates the given expression.
        The generated function accepts the unbound variables in the expression
        as arguments.
    """
    def function(*args, **kwargs):
        """
            A function that evaluates the expression
        """
        if len(args) != 1:
            raise NotImplementedError()

        return expression.execute_direct({'self': args[0]})

    return function


TYPES = {"string": String, "number": Number, "bool": Bool, "list": List, "dict": Dict}



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from enum import Enum


class ResourceState(Enum):
    unavailable = 1  # This state is set by the agent when no handler is available for the resource
    skipped = 2  #
    dry = 3
    deployed = 4
    failed = 5
    deploying = 6
    available = 7
    cancelled = 8  # When a new version is pushed, in progress deploys are cancelled
    undefined = 9  # The state of this resource is unknown at this moment in the orchestration process
    skipped_for_undefined = 10  # This resource depends on an undefined resource
    processing_events = 11


# undeployable
UNDEPLOYABLE_STATES = [ResourceState.undefined, ResourceState.skipped_for_undefined]
# this resource action is not complete, resource is in transient state
TRANSIENT_STATES = [ResourceState.available, ResourceState.deploying, ResourceState.processing_events]
# not counting as done
NOT_DONE_STATES = TRANSIENT_STATES
# counts as done
DONE_STATES = [ResourceState.unavailable, ResourceState.skipped, ResourceState.deployed, ResourceState.failed,
               ResourceState.cancelled] + UNDEPLOYABLE_STATES

# starting states
INITIAL_STATES = [ResourceState.available]
# states one can't transition out off
TERMINAL_STATES = UNDEPLOYABLE_STATES
# states on can transition to
VALID_STATES_ON_STATE_UPDATE = [ResourceState.unavailable, ResourceState.skipped, ResourceState.deployed,
                                ResourceState.failed, ResourceState.deploying, ResourceState.cancelled, ResourceState.undefined,
                                ResourceState.skipped_for_undefined, ResourceState.processing_events]

UNKNOWN_STRING = "<<undefined>>"

"""
States set by server upon upload

1. available
2. undefined (terminal state)
3. skipped_for_undefined (terminal state)

States set by agent
1. skipped
2. failed
3. deployed
4. unavailable
5. cancelled
6. deploying
7. processing_events

Each deploy sets the agent state again, all agent states can transition to all agent states

States that are in the action log, but not actual states
1. dry



                                           +-----> deploying        -<--------+
                                           |                                  |
                                           +-----> processing_events-<--------+
                                           |                                  |
                                           +-----> skipped          -<--------+
                                           |                                  |
                                           +-----> failed           -<--------+
                                           |                                  |
    +---------------->  available  +-------------> unavailable      -<--------+
    |                                      |                                  |
    |                                      +-----> deployed         -<--------+
    |                                      |                                  |
+---+---------+                            +-----> cancelled        -<--------+
| compiler    +------> undefined
+---+---------+
    |
    |
    +----------------> skipped_for_undefined

"""


class Change(Enum):
    nochange = 0
    created = 1
    purged = 2
    updated = 3


class VersionState(Enum):
    success = 1
    failed = 2
    deploying = 3
    pending = 4


class ResourceAction(Enum):
    store = 1
    push = 2
    pull = 3
    deploy = 4
    dryrun = 5
    getfact = 6
    other = 8


STATE_UPDATE = [ResourceAction.deploy]


class AgentTriggerMethod(Enum):
    push_incremental_deploy = 1
    push_full_deploy = 2

    @classmethod
    def get_agent_trigger_method(cls, is_full_deploy):
        if is_full_deploy:
            return cls.push_full_deploy
        else:
            return cls.push_incremental_deploy


class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    TRACE = 3
    NOTSET = 0


INMANTA_URN = "urn:inmanta:"


class Compilestate(Enum):
    success = 1
    failed = 2


EXPORT_META_DATA = "export_metadata"
META_DATA_COMPILE_STATE = "inmanta:compile:state"
INMANTA_MT_HEADER = "X-Inmanta-tid"
VALID_CLIENT_TYPES = ["api", "agent", "compiler", "public"]

# For testing

# assume we are running in a tty
ENVIRON_FORCE_TTY = "FORCE_TTY"


LOG_LEVEL_TRACE = 3

NAME_RESOURCE_ACTION_LOGGER = "resource_action_logger"


# Time we give the server/agent to shutdown gracefully, before we force stop the ioloop
SHUTDOWN_GRACE_IOLOOP = 10
# Time we give the server/agent to shutdown gracefully, before we execute sys.exit(3)
SHUTDOWN_GRACE_HARD = 15
# Hard shutdown exit code
EXIT_HARD = 3

TIME_ISOFMT = "%Y-%m-%dT%H:%M:%S.%f"
TIME_LOGFMT = "%Y-%m-%d %H:%M:%S"

PLUGINS_PACKAGE = "inmanta_plugins"



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""


"""
This module enables tracking of object construction.

Tracker object are attached to all created Instances in the field trackers. Every Instance can have one or more Trackers.

If the Tracker object is a ModuleTracker, the object is created at module level
If the Tracker object is an ImplementsTracker, the object is created in an Implementation block

"""


class Tracker(object):

    pass


class ModuleTracker(Tracker):

    def __init__(self, block):
        self.block = block
        self.namespace = block.namespace

    def get_next(self):
        return []


class ImplementsTracker(Tracker):

    def __init__(self, subc, instance):
        self.instance = instance
        self.subc = subc
        self.implements = subc.implements
        self.implementations = self.implements.implementations

    def get_next(self):
        return self.instance.trackers



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import inmanta.protocol.endpoints
from inmanta.types import JsonType, NoneGen
from inmanta.util import Scheduler
from inmanta.protocol import Client, handle, methods
from inmanta.protocol import common, endpoints
from inmanta.protocol.rest import server

from inmanta import config as inmanta_config
from inmanta.server import config as opt, SLICE_SESSION_MANAGER

from tornado import gen, queues, web, routing
from tornado.ioloop import IOLoop

from typing import Dict, Tuple, Callable, Optional, List, Union

import logging
import asyncio
import time
import uuid
import abc
from asyncio.tasks import ensure_future


LOGGER = logging.getLogger(__name__)


class ReturnClient(Client):
    """
        A client that uses a return channel to connect to its destination. This client is used by the server to communicate
        back to clients over the heartbeat channel.
    """

    def __init__(self, name: str, session: "Session") -> None:
        super().__init__(name)
        self.session = session

    @gen.coroutine
    def _call(self, method_properties: common.MethodProperties, args, kwargs) -> common.Result:
        call_spec = method_properties.build_call(args, kwargs)
        try:
            return_value = yield self.session.put_call(call_spec, timeout=method_properties.timeout)
        except gen.TimeoutError:
            return common.Result(code=500, result={"message": "Call timed out"})

        return common.Result(code=return_value["code"], result=return_value["result"])


# Server Side
class Server(endpoints.Endpoint):
    def __init__(self, connection_timout: int = 120) -> None:
        super().__init__("server")

        self._slices: Dict[str, ServerSlice] = {}
        self._handlers: List[routing.Rule] = []
        self.token: Optional[str] = inmanta_config.Config.get(self.id, "token", None)
        self.connection_timout = connection_timout
        self.sessions_handler = SessionManager()
        self.add_slice(self.sessions_handler)

        self._transport = server.RESTServer(self.sessions_handler, self.id)
        self.running = False

    def add_slice(self, slice: "ServerSlice") -> None:
        """
            Add new endpoints to this rest transport
        """
        self._slices[slice.name] = slice

    def get_slices(self) -> Dict[str, "ServerSlice"]:
        return self._slices

    def get_slice(self, name: str) -> "ServerSlice":
        return self._slices[name]

    def get_id(self) -> str:
        """
            Returns a unique id for a transport on an endpoint
        """
        return "server_rest_transport"

    id = property(get_id)

    @gen.coroutine
    def start(self) -> NoneGen:
        """
            Start the transport.

            The order in which the different endpoints are prestarted/started, is determined by the
            order in which they are added to the RESTserver via the add_endpoint(endpoint) method.
            This order is hardcoded in the get_server_slices() method in server/bootloader.py
        """
        if self.running:
            return
        LOGGER.debug("Starting Server Rest Endpoint")
        self.running = True

        for slice in self.get_slices().values():
            yield slice.prestart(self)

        for slice in self.get_slices().values():
            yield slice.start()
            self._handlers.extend(slice.get_handlers())

        yield self._transport.start(self.get_slices().values(), self._handlers)

    @gen.coroutine
    def stop(self) -> NoneGen:
        """
            Stop the transport.

            The order in which the endpoint are stopped, is reverse compared to the starting order.
            This prevents database connection from being closed too early. This order in which the endpoint
            are started, is hardcoded in the get_server_slices() method in server/bootloader.py
        """
        if not self.running:
            return
        self.running = False
        LOGGER.debug("Stopping Server Rest Endpoint")
        yield self._transport.stop()
        for endpoint in reversed(list(self.get_slices().values())):
            yield endpoint.stop()
        yield self._transport.join()


class ServerSlice(inmanta.protocol.endpoints.CallTarget):
    """
        An API serving part of the server.
    """

    def __init__(self, name: str) -> None:
        super().__init__()

        self._name: str = name
        self._handlers: List[routing.Rule] = []
        self._sched = Scheduler("server slice")  # FIXME: why has each slice its own scheduler?
        self.running: bool = False  # for debugging

    @abc.abstractmethod
    @gen.coroutine
    def prestart(self, server: Server) -> NoneGen:
        """Called by the RestServer host prior to start, can be used to collect references to other server slices"""

    @gen.coroutine
    @abc.abstractmethod
    def start(self) -> NoneGen:
        """
            Start the server slice.
        """
        self.running = True

    @gen.coroutine
    def stop(self) -> NoneGen:
        self.running = False
        self._sched.stop()

    name = property(lambda self: self._name)

    def get_handlers(self) -> List[routing.Rule]:
        return self._handlers

    def add_future(self, future: asyncio.Future) -> None:
        """
            Add a future to the ioloop to be handled, but do not require the result.
        """

        def handle_result(f: asyncio.Future) -> None:
            try:
                f.result()
            except Exception as e:
                LOGGER.exception("An exception occurred while handling a future: %s", str(e))

        IOLoop.current().add_future(ensure_future(future), handle_result)

    def schedule(self, call: Callable, interval: int = 60) -> None:
        self._sched.add_action(call, interval)

    def add_static_handler(self, location: str, path: str, default_filename: Optional[str] = None, start: bool = False) -> None:
        """
            Configure a static handler to serve data from the specified path.
        """
        if location[0] != "/":
            location = "/" + location

        if location[-1] != "/":
            location = location + "/"

        options = {"path": path}
        if default_filename is None:
            options["default_filename"] = "index.html"

        self._handlers.append(routing.Rule(routing.PathMatches(r"%s(.*)" % location), web.StaticFileHandler, options))
        self._handlers.append(routing.Rule(routing.PathMatches(r"%s" % location[:-1]), web.RedirectHandler, {"url": location}))

        if start:
            self._handlers.append((r"/", web.RedirectHandler, {"url": location}))

    def add_static_content(self, path: str, content: str, content_type: str = "application/javascript") -> None:
        self._handlers.append(
            routing.Rule(
                routing.PathMatches(r"%s(.*)" % path),
                server.StaticContentHandler,
                {"transport": self, "content": content, "content_type": content_type},
            )
        )


class Session(object):
    """
        An environment that segments agents connected to the server
    """

    def __init__(
        self,
        sessionstore: "SessionManager",
        sid: uuid.UUID,
        hang_interval: int,
        timout: int,
        tid: uuid.UUID,
        endpoint_names: List[str],
        nodename: str,
    ) -> None:
        self._sid = sid
        self._interval = hang_interval
        self._timeout = timout
        self._sessionstore: SessionManager = sessionstore
        self._seen: float = time.time()
        self._callhandle = None
        self.expired: bool = False

        self.tid: uuid.UUID = tid
        self.endpoint_names: List[str] = endpoint_names
        self.nodename: str = nodename

        self._replies: Dict[uuid.UUID, asyncio.Future] = {}
        self.check_expire()
        self._queue: queues.Queue[common.Request] = queues.Queue()

        self.client = ReturnClient(str(sid), self)

    def check_expire(self) -> None:
        if self.expired:
            LOGGER.exception("Tried to expire session already expired")
        ttw = self._timeout + self._seen - time.time()
        if ttw < 0:
            self.expire(self._seen - time.time())
        else:
            self._callhandle = IOLoop.current().call_later(ttw, self.check_expire)

    def get_id(self) -> uuid.UUID:
        return self._sid

    id = property(get_id)

    def expire(self, timeout: float) -> None:
        self.expired = True
        if self._callhandle is not None:
            IOLoop.current().remove_timeout(self._callhandle)
        self._sessionstore.expire(self, timeout)

    def seen(self) -> None:
        self._seen = time.time()

    def _set_timeout(self, future: asyncio.Future, timeout: int, log_message: str) -> None:
        def on_timeout():
            if not self.expired:
                LOGGER.warning(log_message)
            future.set_exception(gen.TimeoutError())

        timeout_handle = IOLoop.current().add_timeout(IOLoop.current().time() + timeout, on_timeout)
        future.add_done_callback(lambda _: IOLoop.current().remove_timeout(timeout_handle))

    def put_call(self, call_spec: common.Request, timeout: int = 10) -> asyncio.Future:
        future = asyncio.Future()

        reply_id = uuid.uuid4()

        LOGGER.debug("Putting call %s: %s %s for agent %s in queue", reply_id, call_spec.method, call_spec.url, self._sid)

        call_spec.reply_id = reply_id
        self._queue.put(call_spec)
        self._set_timeout(
            future, timeout, "Call %s: %s %s for agent %s timed out." % (reply_id, call_spec.method, call_spec.url, self._sid)
        )
        self._replies[reply_id] = future

        return future

    @gen.coroutine
    def get_calls(self) -> Optional[List[common.Request]]:
        """
            Get all calls queued for a node. If no work is available, wait until timeout. This method returns none if a call
            fails.
        """
        try:
            call_list: List[common.Request] = []
            call = yield self._queue.get(timeout=IOLoop.current().time() + self._interval)
            if call is None:
                # aborting session
                return None
            call_list.append(call)
            while self._queue.qsize() > 0:
                call = yield self._queue.get()
                if call is None:
                    # aborting session
                    return None
                call_list.append(call)

            return call_list

        except gen.TimeoutError:
            return None

    def set_reply(self, reply_id: uuid.UUID, data: JsonType) -> None:
        LOGGER.log(3, "Received Reply: %s", reply_id)
        if reply_id in self._replies:
            future: asyncio.Future = self._replies[reply_id]
            del self._replies[reply_id]
            if not future.done():
                future.set_result(data)
        else:
            LOGGER.debug("Received Reply that is unknown: %s", reply_id)

    def get_client(self) -> ReturnClient:
        return self.client

    def abort(self):
        "Send poison pill to signal termination."
        self._queue.put(None)


class SessionListener(object):
    def new_session(self, session: Session) -> None:
        pass

    def expire(self, session: Session, timeout: float) -> None:
        pass

    def seen(self, session: Session, endpoint_names: List[str]) -> None:
        pass


# Internals
class SessionManager(ServerSlice):
    """
        A service that receives method calls over one or more transports
    """

    __methods__: Dict[str, Tuple[str, Callable]] = {}

    def __init__(self) -> None:
        super().__init__(SLICE_SESSION_MANAGER)

        # Config
        interval: int = opt.agent_timeout.get()
        hangtime: Optional[int] = opt.agent_hangtime.get()

        if hangtime is None:
            hangtime = int(interval * 3 / 4)

        self.hangtime: int = hangtime
        self.interval: int = interval

        # Session management
        self._sessions: Dict[uuid.UUID, Session] = {}

        # Listeners
        self.listeners: List[SessionListener] = []

    def add_listener(self, listener: SessionListener) -> None:
        self.listeners.append(listener)

    @gen.coroutine
    def prestart(self, server: Server) -> None:
        """Called by the RestServer host prior to start, can be used to collect references to other server slices"""

    @gen.coroutine
    def start(self) -> None:
        """
            Start the server slice.
        """
        yield super().start()

    @gen.coroutine
    def stop(self) -> None:
        """
            Stop the end-point and all of its transports
        """
        yield super().stop()
        # terminate all sessions cleanly
        for session in self._sessions.copy().values():
            session.expire(0)
            session.abort()

    def validate_sid(self, sid: uuid.UUID) -> bool:
        if isinstance(sid, str):
            sid = uuid.UUID(sid)
        return sid in self._sessions

    def get_or_create_session(self, sid: uuid.UUID, tid: uuid.UUID, endpoint_names: List[str], nodename: str) -> Session:
        if isinstance(sid, str):
            sid = uuid.UUID(sid)

        if sid not in self._sessions:
            session = self.new_session(sid, tid, endpoint_names, nodename)
            self._sessions[sid] = session
            for listener in self.listeners:
                listener.new_session(session)
        else:
            session = self._sessions[sid]
            self.seen(session, endpoint_names)
            for listener in self.listeners:
                listener.seen(session, endpoint_names)

        return session

    def new_session(self, sid: uuid.UUID, tid: uuid.UUID, endpoint_names: List[str], nodename: str) -> Session:
        LOGGER.debug("New session with id %s on node %s for env %s with endpoints %s" % (sid, nodename, tid, endpoint_names))
        return Session(self, sid, self.hangtime, self.interval, tid, endpoint_names, nodename)

    def expire(self, session: Session, timeout: float) -> None:
        LOGGER.debug("Expired session with id %s, last seen %d seconds ago" % (session.get_id(), timeout))
        for listener in self.listeners:
            listener.expire(session, timeout)
        del self._sessions[session.id]

    def seen(self, session: Session, endpoint_names: List[str]) -> None:
        LOGGER.debug("Seen session with id %s" % (session.get_id()))
        session.seen()

    @handle(methods.heartbeat, env="tid")
    @gen.coroutine
    def heartbeat(
        self, sid: uuid.UUID, env: "inmanta.data.Environment", endpoint_names, nodename
    ) -> Union[int, Tuple[int, Dict[str, str]]]:
        LOGGER.debug("Received heartbeat from %s for agents %s in %s", nodename, ",".join(endpoint_names), env.id)

        session: Session = self.get_or_create_session(sid, env.id, endpoint_names, nodename)

        LOGGER.debug("Let node %s wait for method calls to become available. (long poll)", nodename)
        call_list = yield session.get_calls()
        if call_list is not None:
            LOGGER.debug("Pushing %d method calls to node %s", len(call_list), nodename)
            return 200, {"method_calls": call_list}
        else:
            LOGGER.debug("Heartbeat wait expired for %s, returning. (long poll)", nodename)

        return 200

    @handle(methods.heartbeat_reply)
    @gen.coroutine
    def heartbeat_reply(
        self, sid: uuid.UUID, reply_id: uuid.UUID, data: JsonType
    ) -> Union[int, Tuple[int, Dict[str, str]]]:
        try:
            env = self._sessions[sid]
            env.set_reply(reply_id, data)
            return 200
        except Exception:
            LOGGER.warning("could not deliver agent reply with sid=%s and reply_id=%s" % (sid, reply_id), exc_info=True)
            return 500



# Adapted from pyformance

import asyncio
import time
import base64
import logging
from asyncio import Task

from pyformance import global_registry, MetricsRegistry
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
from urllib.parse import quote

from typing import Optional, Dict

LOGGER = logging.getLogger(__name__)

DEFAULT_INFLUX_SERVER = "127.0.0.1"
DEFAULT_INFLUX_PORT = 8086
DEFAULT_INFLUX_DATABASE = "metrics"
DEFAULT_INFLUX_USERNAME = None
DEFAULT_INFLUX_PASSWORD = None
DEFAULT_INFLUX_PROTOCOL = "http"


class AsyncReporter(object):
    def __init__(
        self,
        registry: MetricsRegistry = None,
        reporting_interval: int = 30
    ) -> None:
        self.registry = registry or global_registry()
        self.reporting_interval = reporting_interval
        self._stopped = False
        self._handle: Optional[Task[None]] = None

    def start(self) -> bool:
        if self._stopped:
            return False
        self._handle = asyncio.get_event_loop().create_task(self._loop())
        return True

    def stop(self) -> None:
        self._stopped = True
        if self._handle is not None:
            self._handle.cancel()

    async def _loop(self) -> None:
        loop = asyncio.get_event_loop()
        next_loop_time = loop.time()
        while not self._stopped:
            try:
                await self.report_now(self.registry)
            except Exception:
                LOGGER.warning("Could not send metrics report", exc_info=True)
            next_loop_time += self.reporting_interval
            wait = max(0, next_loop_time - loop.time())
            await asyncio.sleep(wait)

    async def report_now(
        self, registry: MetricsRegistry = None, timestamp: float = None
    ) -> None:
        raise NotImplementedError(self.report_now)


class InfluxReporter(AsyncReporter):

    """
    InfluxDB reporter using native http api
    (based on https://influxdb.com/docs/v1.1/guides/writing_data.html)
    """

    def __init__(
        self,
        registry: Optional[MetricsRegistry] = None,
        reporting_interval: int = 5,
        database: str = DEFAULT_INFLUX_DATABASE,
        server: str = DEFAULT_INFLUX_SERVER,
        username: Optional[str] = DEFAULT_INFLUX_USERNAME,
        password: Optional[str] = DEFAULT_INFLUX_PASSWORD,
        port: int = DEFAULT_INFLUX_PORT,
        protocol: str = DEFAULT_INFLUX_PROTOCOL,
        autocreate_database: bool = False,
        tags: Dict[str, str] = {},
    ) -> None:
        super(InfluxReporter, self).__init__(registry, reporting_interval)
        self.database = database
        self.username = username
        self.password = password
        self.port = port
        self.protocol = protocol
        self.server = server
        self.autocreate_database = autocreate_database
        self._did_create_database = False
        self.tags = tags
        self.key = "metrics"
        if self.tags:
            tagstring = ",".join(
                "%s=%s" % (key, value) for key, value in self.tags.items()
            )
            self.key = "%s,%s" % (self.key, tagstring)
        self.key = "%s,key=" % self.key

    async def _create_database(self, http_client: AsyncHTTPClient) -> None:
        url = "%s://%s:%s/query" % (self.protocol, self.server, self.port)
        q = quote("CREATE DATABASE %s" % self.database)
        request = HTTPRequest(url + "?q=" + q)
        if self.username and self.password:
            auth = _encode_username(self.username, self.password)
            request.headers.add("Authorization", "Basic %s" % auth.decode("utf-8"))
        try:
            response = await http_client.fetch(request)
            response.rethrow()
            # Only set if we actually were able to get a successful response
            self._did_create_database = True
        except Exception:
            LOGGER.warning(
                "Cannot create database %s to %s",
                self.database,
                self.server,
                exc_info=True
            )

    async def report_now(
        self,
        registry: Optional[MetricsRegistry] = None,
        timestamp: Optional[float] = None,
    ):
        http_client = AsyncHTTPClient()

        if self.autocreate_database and not self._did_create_database:
            await self._create_database(http_client)
        timestamp = timestamp or int(round(time.time()))
        metrics = (registry or self.registry).dump_metrics()
        post_data = []
        for key, metric_values in metrics.items():
            table = self.key + key
            values = ",".join(
                [
                    "%s=%s" % (k, v if type(v) is not str else '"{}"'.format(v))
                    for (k, v) in metric_values.items()
                ]
            )
            line = "%s %s %s" % (table, values, timestamp)
            post_data.append(line)
        post_data_all = "\n".join(post_data)
        path = "/write?db=%s&precision=s" % self.database
        url = "%s://%s:%s%s" % (self.protocol, self.server, self.port, path)
        request = HTTPRequest(url, method="POST", body=post_data_all.encode("utf-8"))
        if self.username and self.password:
            auth = _encode_username(self.username, self.password)
            request.headers.add("Authorization", "Basic %s" % auth.decode("utf-8"))
        try:
            response = await http_client.fetch(request)
            response.rethrow()
        except HTTPError:
            LOGGER.warning("Cannot write to %s", self.server, exc_info=True)


def _encode_username(username: str, password: str) -> bytes:
    auth_string = ("%s:%s" % (username, password)).encode()
    return base64.b64encode(auth_string)



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import re

from inmanta.ast import Namespace, LocatableString
from inmanta.ast.statements import define, Literal
from inmanta.parser.plyInmantaParser import parse
from inmanta.parser import ParserException
from inmanta.ast.statements.define import (
    DefineImplement,
    DefineTypeConstraint,
    DefineTypeDefault,
    DefineIndex,
    DefineEntity,
    DefineImplementInherits,
)
from inmanta.ast.constraint.expression import (
    GreaterThan,
    Regex,
    Not,
    And,
    IsDefined,
    In,
)
from inmanta.ast.statements.generator import Constructor
from inmanta.ast.statements.call import FunctionCall
from inmanta.ast.statements.assign import (
    Assign,
    CreateList,
    IndexLookup,
    StringFormat,
    CreateDict,
    ShortIndexLookup,
    SetAttribute,
    MapLookup,
)
from inmanta.ast.variables import Reference, AttributeReference
import pytest
from inmanta.execute.util import NoneValue


def parse_code(model_code: str):
    root_ns = Namespace("__root__")
    main_ns = Namespace("__config__")
    main_ns.parent = root_ns
    statements = parse(main_ns, "test", model_code)

    return statements


def test_define_empty():
    parse_code("""""")


def test_define_entity():
    """Test the definition of entities
    """
    statements = parse_code(
        """
entity Test:
end
entity Other:
string hello
end
entity Other:
 \"\"\"XX
 \"\"\"
end
"""
    )

    assert len(statements) == 3

    stmt = statements[0]
    assert isinstance(stmt, define.DefineEntity)
    assert stmt.name == "Test"
    assert stmt.parents == ["std::Entity"]
    assert len(stmt.attributes) == 0
    assert stmt.comment is None


def test_undefine_default():
    statements = parse_code(
        """
entity Test extends Foo:
 string hello = undef
 string[] dinges = undef
end"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert isinstance(stmt, define.DefineEntity)
    assert stmt.name == "Test"
    assert stmt.parents == ["Foo"]
    assert len(stmt.attributes) == 2
    assert stmt.comment is None

    for ad in stmt.attributes:
        assert isinstance(ad.type, LocatableString)
        assert isinstance(ad.name, LocatableString)
        assert ad.default is None
        assert ad.remove_default

    assert str(stmt.attributes[0].name) == "hello"
    assert str(stmt.attributes[1].name) == "dinges"


def test_extend_entity():
    """Test extending entities
    """
    statements = parse_code(
        """
entity Test extends Foo:
end
"""
    )

    assert len(statements) == 1

    stmt = statements[0]
    assert stmt.parents == ["Foo"]


def test_complex_entity():
    """Test definition of a complex entity
    """
    documentation = "This entity has documentation"
    statements = parse_code(
        """
entity Test extends Foo, foo::sub::Bar:
    \"\"\" %s
    \"\"\"
    string hello
    bool bar = true
    number? ten=5
end
"""
        % documentation
    )

    assert len(statements) == 1

    stmt = statements[0]
    assert len(stmt.parents) == 2
    assert stmt.parents == ["Foo", "foo::sub::Bar"]
    assert str(stmt.comment).strip() == documentation
    assert len(stmt.attributes) == 3

    for ad in stmt.attributes:
        assert isinstance(ad.type, LocatableString)
        assert isinstance(ad.name, LocatableString)

    assert str(stmt.attributes[0].name) == "hello"
    assert str(stmt.attributes[1].name) == "bar"
    assert str(stmt.attributes[2].name) == "ten"

    assert stmt.attributes[1].default.execute(None, None, None)

    assert stmt.attributes[2].default.execute(None, None, None) == 5


def test_relation():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test tests [0:] -- [5:10] Foo bars
"""
    )

    assert len(statements) == 1
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bars"

    assert rel.left[2] == (0, None)
    assert rel.right[2] == (5, 10)


def test_relation_2():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test tests [3] -- [:10] Foo bars
"""
    )

    assert len(statements) == 1
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bars"

    assert rel.left[2] == (3, 3)
    assert rel.right[2] == (None, 10)


def test_new_relation():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] -- Foo.tests [5:10]
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] == (5, 10)
    assert rel.right[2] == (1, 1)


def test_new_relation_with_annotations():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] foo,bar Foo.tests [5:10]
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] == (5, 10)
    assert rel.right[2] == (1, 1)
    assert len(rel.annotations) == 2
    assert rel.annotation_expression[0][1].name == "foo"
    assert rel.annotation_expression[1][1].name == "bar"


def test_new_relation_unidir():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] -- Foo
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert (rel.left[1]) is None
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] is None
    assert rel.right[2] == (1, 1)


def test_new_relation_with_annotations_unidir():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] foo,bar Foo
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert rel.left[1] is None
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] is None
    assert rel.right[2] == (1, 1)
    assert len(rel.annotations) == 2
    assert rel.annotation_expression[0][1].name == "foo"
    assert rel.annotation_expression[1][1].name == "bar"


def test_implementation():
    """Test the definition of implementations
    """
    statements = parse_code(
        """
implementation test for Test:
end
"""
    )

    assert len(statements) == 1
    assert len(statements[0].block.get_stmts()) == 0
    assert statements[0].name == "test"
    assert isinstance(statements[0].entity, str)

    statements = parse_code(
        """
implementation test for Test:
    std::File(attr="a")
    var = hello::func("world")
end
"""
    )

    assert len(statements) == 1
    assert len(statements[0].block.get_stmts()) == 2


def test_implementation_with_for():
    """Test the propagation of type requires when using a for
    """
    statements = parse_code(
        """
implementation test for Test:
    \"\"\" test \"\"\"
    for v in data:
        std::template("template")
    end
end
"""
    )

    assert len(statements) == 1
    assert len(statements[0].block.get_stmts()) == 1


def test_implements():
    """Test implements with no selector
    """
    statements = parse_code(
        """
implement Test using test
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert str(stmt.entity) == "Test"
    assert stmt.implementations == ["test"]
    assert str(stmt.select) == "True"


def test_implements_2():
    """Test implements with selector
    """
    statements = parse_code(
        """
implement Test using test, blah when (self > 5)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert str(stmt.entity) == "Test"
    assert stmt.implementations == ["test", "blah"]
    assert isinstance(stmt.select, GreaterThan)
    assert stmt.select.children[0].name == "self"
    assert stmt.select.children[1].value == 5


def test_implements_parent():
    statements = parse_code(
        """
implement Test using parents  \""" testc \"""
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplementInherits)
    assert stmt.entity == "Test"


def test_implements_selector():
    """Test implements with selector
    """
    statements = parse_code(
        """
implement Test using test when not (fg(self) and false)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert stmt.entity == "Test"
    assert stmt.implementations == ["test"]
    assert isinstance(stmt.select, Not)
    assert isinstance(stmt.select.children[0], And)
    assert isinstance(stmt.select.children[0].children[0], FunctionCall)
    assert isinstance(stmt.select.children[0].children[1], Literal)


def test_regex():
    statements = parse_code(
        """
a = /[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}/
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, Regex)
    assert stmt.children[1].value == re.compile(
        r"[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}"
    )


def test_regex_escape():
    statements = parse_code(
        r"""
a = /\/1/
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, Regex)
    assert stmt.children[1].value == re.compile(r"\/1")


def test_regex_twice():
    statements = parse_code(
        r"""
a = /\/1/
b = "v"
c = /\/1/
"""
    )

    assert len(statements) == 3
    stmt = statements[0].value
    assert isinstance(stmt, Regex)
    assert stmt.children[1].value == re.compile(r"\/1")


def test_typedef():
    statements = parse_code(
        """
typedef uuid as string matching /[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}/
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineTypeConstraint)
    assert str(stmt.name) == "uuid"
    assert stmt.basetype == "string"
    assert isinstance(stmt.get_expression(), Regex)
    assert stmt.get_expression().children[1].value == re.compile(
        r"[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}"
    )


def test_typedef_in():
    statements = parse_code(
        """
typedef abc as string matching self in ["a","b","c"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineTypeConstraint)
    assert str(stmt.name) == "abc"
    assert stmt.basetype == "string"
    assert isinstance(stmt.get_expression(), In)
    assert [x.value for x in stmt.get_expression().children[1].items] == ["a", "b", "c"]


def test_typedef2():
    statements = parse_code(
        """
typedef ConfigFile as File(mode = 644, owner = "root", group = "root")
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineTypeDefault)
    assert stmt.name == "ConfigFile"
    assert isinstance(stmt.ctor, Constructor)


def test_index():
    statements = parse_code(
        """
index File(host, path)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineIndex)
    assert stmt.type == "File"
    assert stmt.attributes == ["host", "path"]


def test_ctr():
    statements = parse_code(
        """
File(host = 5, path = "Jos")
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Constructor)
    assert str(stmt.class_type) == "File"
    assert {k: v.value for k, v in stmt.attributes.items()} == {
        "host": 5,
        "path": "Jos",
    }


def test_indexlookup():
    statements = parse_code(
        """
a=File[host = 5, path = "Jos"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, IndexLookup)
    assert stmt.index_type == "File"
    assert {k: v.value for k, v in stmt.query} == {"host": 5, "path": "Jos"}


def test_short_index_lookup():
    statements = parse_code(
        """
a = vm.files[path="/etc/motd"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, ShortIndexLookup)
    assert isinstance(stmt.rootobject, Reference)
    assert stmt.rootobject.name == "vm"
    assert stmt.relation == "files"
    assert {k: v.value for k, v in stmt.querypart} == {"path": "/etc/motd"}


def test_ctr_2():
    statements = parse_code(
        """
File( )
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Constructor)
    assert str(stmt.class_type) == "File"
    assert {k: v.value for k, v in stmt.attributes.items()} == {}


def test_function():
    statements = parse_code(
        """
file( )
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, FunctionCall)
    assert stmt.name == "file"


def test_function_2():
    statements = parse_code(
        """
file(b)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, FunctionCall)
    assert stmt.name == "file"


def test_function_3():
    statements = parse_code(
        """
file(b,)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, FunctionCall)
    assert stmt.name == "file"


def test_list_def():
    statements = parse_code(
        """
a=["a]","b"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateList)
    assert [x.value for x in stmt.value.items] == ["a]", "b"]


def test_list_def_trailing_comma():
    statements = parse_code(
        """
a=["a]","b",]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateList)
    assert [x.value for x in stmt.value.items] == ["a]", "b"]


def test_map_def():
    statements = parse_code(
        """
a={ "a":"b", "b":1}
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert [(x[0], x[1].value) for x in stmt.value.items] == [("a", "b"), ("b", 1)]


def test_map_def_var():
    statements = parse_code("""a={ "b":b}""")
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert isinstance(stmt.value.items[0][1], Reference)


def test_map_def_list():
    statements = parse_code("""a={ "a":["a"]}""")
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert isinstance(stmt.value.items[0][1], CreateList)


def test_map_def_map():
    statements = parse_code("""a={ "a":{"a":"C"}}""")
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert isinstance(stmt.value.items[0][1], CreateDict)


def test_booleans():
    statements = parse_code(
        """
a=true b=false
"""
    )

    assert len(statements) == 2
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert stmt.value.value
    assert not statements[1].value.value


def test_none():
    statements = parse_code(
        """
a=null
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value.value, NoneValue)


def test_numbers():
    statements = parse_code(
        """
a=1
b=2.0
c=-5
d=-0.256
"""
    )

    assert len(statements) == 4
    values = [1, 2.0, -5, -0.256]
    for i in range(4):
        stmt = statements[i]
        assert isinstance(stmt, Assign)
        assert stmt.value.value == values[i]


def test_string():
    statements = parse_code(
        """
a="jos"
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == "jos"


def test_string_2():
    statements = parse_code(
        """
a='jos'
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == "jos"


def test_empty():
    statements = parse_code(
        """
a=""
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == ""


def test_empty_2():
    statements = parse_code(
        """
a=''
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == ""


def test_string_format():
    statements = parse_code(
        """
a="j{{o}}s"
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, StringFormat)
    assert isinstance(stmt.value._variables[0][0], Reference)
    assert [x[0].name for x in stmt.value._variables] == ["o"]


def test_string_format_2():
    statements = parse_code(
        """
a="j{{c.d}}s"
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, StringFormat)
    assert len(stmt.value._variables) == 1
    assert len(stmt.value._variables[0]) == 2
    assert isinstance(stmt.value._variables[0][0], AttributeReference)
    assert stmt.value._variables[0][0].instance.name == "c"
    assert stmt.value._variables[0][0].attribute == "d"


def test_attribute_reference():
    statements = parse_code(
        """
a=a::b::c.d
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, AttributeReference)
    assert isinstance(stmt.value.instance, Reference)
    assert stmt.value.instance.full_name == "a::b::c"
    assert stmt.value.attribute == "d"


def test_is_defined():
    statements = parse_code(
        """
implement Test1 using tt when self.other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert stmt.select.attr.name == "self"
    assert stmt.select.name == "other"


def test_is_defined_implicit_self():
    statements = parse_code(
        """
implement Test1 using tt when other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert stmt.select.attr.name == "self"
    assert stmt.select.name == "other"


def test_is_defined_short():
    statements = parse_code(
        """
implement Test1 using tt when a.other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert isinstance(stmt.select.attr, AttributeReference)
    assert stmt.select.attr.instance.name == "self"
    assert stmt.select.attr.attribute == "a"
    assert stmt.select.name == "other"


def assert_is_non_value(x):
    assert isinstance(x, Literal)
    assert isinstance(x.value, NoneValue)


def compare_attr(attr, name, mytype, defs, multi=False, opt=False):
    assert str(attr.name) == name
    defs(attr.default)
    assert attr.multi == multi
    assert str(attr.type) == mytype
    assert attr.nullable == opt


def assert_is_none(x):
    assert x is None


def assert_equals(x, y):
    assert x == y


def test_define_list_attribute():
    statements = parse_code(
        """
entity Jos:
  bool[] bar
  ip::ip[] ips = ["a"]
  string[] floom = []
  string[] floomx = ["a", "b"]
  string[]? floomopt = null
end"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineEntity)
    assert len(stmt.attributes) == 5

    compare_attr(stmt.attributes[0], "bar", "bool", assert_is_none, multi=True)
    compare_attr(
        stmt.attributes[2],
        "floom",
        "string",
        lambda x: assert_equals([], x.items),
        multi=True,
    )

    def compare_default(list):
        def comp(x):
            assert len(list) == len(x.items)
            for one, it in zip(list, x.items):
                assert isinstance(it, Literal)
                assert it.value == one

        return comp

    compare_attr(
        stmt.attributes[1], "ips", "ip::ip", compare_default(["a"]), multi=True
    )
    compare_attr(
        stmt.attributes[3], "floomx", "string", compare_default(["a", "b"]), multi=True
    )
    compare_attr(
        stmt.attributes[4],
        "floomopt",
        "string",
        assert_is_non_value,
        opt=True,
        multi=True,
    )


def test_define_dict_attribute():
    statements = parse_code(
        """
entity Jos:
  dict bar
  dict foo = {}
  dict blah = {"a":"a"}
  dict? xxx = {"a":"a"}
  dict? xxxx = null
end"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineEntity)
    assert len(stmt.attributes) == 5

    compare_attr(stmt.attributes[0], "bar", "dict", assert_is_none)
    compare_attr(
        stmt.attributes[1], "foo", "dict", lambda x: assert_equals([], x.items)
    )

    def compare_default(list):
        def comp(x):
            assert len(list) == len(x.items)
            for (ok, ov), (k, v) in zip(list, x.items):
                assert k == ok
                assert ov == v.value

        return comp

    compare_attr(stmt.attributes[2], "blah", "dict", compare_default([("a", "a")]))
    compare_attr(
        stmt.attributes[3], "xxx", "dict", compare_default([("a", "a")]), opt=True
    )
    compare_attr(stmt.attributes[4], "xxxx", "dict", assert_is_non_value, opt=True)


def test_lexer():
    parse_code(
        """
#test
//test2
a=0.5
b=""
"""
    )


def test_eol_comment():
    parse_code(
        """a="a"
    # valid_target_types: tosca.capabilities.network.Bindable"""
    )


def test_mls():
    statements = parse_code(
        """
entity MANO:
    \"""
        This entity provides management, orchestration and monitoring

        More test
    \"""
end
"""
    )
    assert len(statements) == 1
    stmt = statements[0]

    assert isinstance(stmt, DefineEntity)

    mls = stmt.comment

    print(mls)

    assert (
        str(mls)
        == """
        This entity provides management, orchestration and monitoring

        More test
    """
    )


def test_bad():
    with pytest.raises(ParserException):
        parse_code(
            """
a = b.c
a=a::b::c.
"""
        )


def test_bad_2():
    with pytest.raises(ParserException):
        parse_code(
            """
a=|
"""
        )


def test_error_on_relation():
    with pytest.raises(ParserException) as e:
        parse_code(
            """
Host.provider [1] -- Provider test"""
        )
    assert e.value.location.file == "test"
    assert e.value.location.lnr == 3
    assert e.value.location.start_char == 2


def test_doc_string_on_new_relation():
    statements = parse_code(
        """
File.host [1] -- Host
\"""
Each file needs to be associated with a host
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Each file needs to be associated with a host"


def test_doc_string_on_relation():
    statements = parse_code(
        """
File file [1] -- [0:] Host host
\"""
Each file needs to be associated with a host
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Each file needs to be associated with a host"


def test_doc_string_on_typedef():
    statements = parse_code(
        """
typedef foo as string matching /^a+$/
\"""
    Foo is a stringtype that only allows "a"
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == 'Foo is a stringtype that only allows "a"'


def test_doc_string_on_typedefault():
    statements = parse_code(
        """
typedef Foo as File(x=5)
\"""
    Foo is a stringtype that only allows "a"
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == 'Foo is a stringtype that only allows "a"'


def test_doc_string_on_impl():
    statements = parse_code(
        """
implementation test for Host:
    \"""
        Bla bla
    \"""
end
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Bla bla"


def test_doc_string_on_implements():
    statements = parse_code(
        """
implement Host using test
\"""
    Always use test!
\"""
\"""
    Not a comment
\"""

"""
    )
    assert len(statements) == 2

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Always use test!"


def test_precise_lexer_positions():
    statements = parse_code(
        """
implement Test1 using tt when self.other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert stmt.select.attr.name == "self"
    assert str(stmt.select.name) == "other"


def test_list_extend_bad():
    with pytest.raises(ParserException):
        parse_code(
            """
    a+=b
    """
        )


def test_list_extend_good():
    statements = parse_code(
        """
z.a+=b
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, SetAttribute)
    assert stmt.list_only is True
    assert isinstance(stmt.value, Reference)
    assert stmt.value.name == "b"


def test_mapref():
    """Test extending entities
    """
    statements = parse_code(
        """
a = b.c["test"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, MapLookup)
    assert isinstance(stmt.value.themap, AttributeReference)
    assert stmt.value.themap.instance.name == "b"
    assert stmt.value.themap.attribute == "c"
    assert isinstance(stmt.value.key, Literal)
    assert stmt.value.key.value == "test"


def test_mapref_2():
    """Test extending entities
    """
    statements = parse_code(
        """
a = c["test"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, MapLookup)
    assert isinstance(stmt.value.themap, Reference)
    assert stmt.value.themap.name == "c"
    assert isinstance(stmt.value.key, Literal)
    assert stmt.value.key.value == "test"


def test_map_multi_ref():
    """Test extending entities
    """
    statements = parse_code(
        """
a = c["test"]["xx"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, MapLookup)
    assert isinstance(stmt.value.themap, MapLookup)
    assert isinstance(stmt.value.themap.themap, Reference)
    assert stmt.value.themap.themap.name == "c"
    assert isinstance(stmt.value.key, Literal)
    assert stmt.value.key.value == "xx"
    assert isinstance(stmt.value.themap.key, Literal)
    assert stmt.value.themap.key.value == "test"



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from unittest.mock import Mock
from uuid import uuid4, UUID
import asyncio

import pytest
from inmanta.server.agentmanager import AgentManager
from inmanta import data
from inmanta.protocol import Result
from utils import assert_equal_ish, UNKWN


class Collector(object):

    def __init__(self):
        self.values = []

    def __call__(self, arg):
        self.values.append(arg)

    async def proccess(self):
        while len(self.values) > 0:
            x = self.values
            self.values = []
            await asyncio.gather(*x)


async def empty_future(*args, **kwargs):
    pass


class MockSession(object):
    """
        An environment that segments agents connected to the server
    """

    def __init__(self, sid, tid, endpoint_names, nodename):
        self._sid = sid
        self.tid = tid
        self.endpoint_names = endpoint_names
        self.nodename = nodename
        self.client = Mock()
        self.client.set_state.side_effect = empty_future

    def get_id(self):
        return self._sid

    id = property(get_id)

    def get_client(self):
        return self.client


@pytest.mark.asyncio(timeout=30)
async def test_primary_selection(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()

    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_future.side_effect = futures
    am = AgentManager(server, False)
    am.add_future = futures
    am.running = True

    async def assert_agent(name: str, state: str, sid: UUID):
        agent = await data.Agent.get(env.id, name)

        assert agent.get_status() == state
        if state == "paused":
            assert agent.primary is None
            assert agent.paused
        elif state == "down":
            assert agent.primary is None
            assert not agent.paused
        elif state == "up":
            assert agent.primary is not None
            agent_instance = await data.AgentInstance.get_by_id(agent.primary)
            agent_proc = await data.AgentProcess.get_one(sid=agent_instance.process)
            assert agent_proc.sid == sid
            assert agent.get_status() == "up"

    async def assert_agents(s1, s2, s3, sid1=None, sid2=None, sid3=None):
        await assert_agent("agent1", s1, sid1)
        await assert_agent("agent2", s2, sid2)
        await assert_agent("agent3", s3, sid3)

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    await futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", True)
    ts1.get_client().reset_mock()
    await assert_agents("paused", "up", "down", sid2=ts1.id)

    # test is_primary
    assert am.is_primary(env, ts1.id, "agent2")
    assert not am.is_primary(env, ts1.id, "agent1")
    assert not am.is_primary(env, uuid4(), "agent2")

    # alive
    am.seen(ts1, ["agent1", "agent2"])
    await futures.proccess()
    assert len(am.sessions) == 1
    await assert_agents("paused", "up", "down", sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    await futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", True)
    ts2.get_client().reset_mock()
    await assert_agents("paused", "up", "up", sid2=ts1.id, sid3=ts2.id)

    # test is_primary
    assert not am.is_primary(env, ts2.id, "agent1")
    assert not am.is_primary(env, ts1.id, "agent1")
    assert am.is_primary(env, ts1.id, "agent2")
    assert am.is_primary(env, ts2.id, "agent3")

    # expire first
    am.expire(ts1, 100)
    await futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", True)
    ts2.get_client().reset_mock()
    await assert_agents("paused", "up", "up", sid2=ts2.id, sid3=ts2.id)

    # expire second
    am.expire(ts2, 100)
    await futures.proccess()
    assert len(am.sessions) == 0
    await assert_agents("paused", "down", "down")

    # test is_primary
    assert not am.is_primary(env, ts1.id, "agent2")
    assert not am.is_primary(env, ts2.id, "agent3")


@pytest.mark.asyncio(timeout=30)
async def test_api(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()
    env2 = data.Environment(name="testenv2", project=project.id)
    await env2.insert()
    env3 = data.Environment(name="testenv3", project=project.id)
    await env3.insert()
    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()
    await data.Agent(environment=env2.id, name="agent4", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_future.side_effect = futures
    am = AgentManager(server, False)
    am.add_future = futures
    am.running = True

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    # third session
    ts3 = MockSession(uuid4(), env3.id, ["agentx"], "ts3")
    am.new_session(ts3)

    await futures.proccess()
    assert len(am.sessions) == 3

    code, all_agents = await am.list_agent_processes(None, None)
    assert code == 200

    shouldbe = {'processes': [{'first_seen': UNKWN, 'expired': None, 'hostname': 'ts1',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent1', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent2', 'process': UNKWN}],
                               'environment': env.id},
                              {'first_seen': UNKWN, 'expired': None, 'hostname': 'ts2',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent2', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent3', 'process': UNKWN}],
                               'environment': env.id},
                              {'first_seen': UNKWN, 'expired': None, 'hostname': 'ts3',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agentx', 'process': UNKWN}],
                               'environment': env3.id}]}

    assert_equal_ish(shouldbe, all_agents, sortby=["hostname", "name"])
    agentid = all_agents['processes'][0]['sid']

    code, all_agents = await am.list_agent_processes(env.id, None)
    assert code == 200

    shouldbe = {'processes': [{'first_seen': UNKWN, 'expired': None, 'hostname': 'ts1',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent1', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent2', 'process': UNKWN}],
                               'environment': env.id},
                              {'first_seen': UNKWN, 'expired': None, 'hostname': 'ts2',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent2', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent3', 'process': UNKWN}],
                               'environment': env.id}]}

    assert_equal_ish(shouldbe, all_agents, sortby=["hostname", "name"])

    code, all_agents = await am.list_agent_processes(env2.id, None)
    assert code == 200

    shouldbe = {'processes': []}

    assert_equal_ish(shouldbe, all_agents)

    async def dummy_status():
        return Result(200, "X")

    ts1.get_client().get_status.side_effect = dummy_status
    report = await am.get_agent_process_report(agentid)
    assert (200, "X") == report

    report = await am.get_agent_process_report(uuid4())
    assert 404 == report[0]

    code, all_agents = await am.list_agents(None)
    assert code == 200
    shouldbe = {'agents': [{'name': 'agent1', 'paused': True, 'last_failover': '', 'primary': '',
                            'environment': env.id, "state": "paused"},
                           {'name': 'agent2', 'paused': False, 'last_failover': UNKWN,
                               'primary': UNKWN, 'environment': env.id, "state": "up"},
                           {'name': 'agent3', 'paused': False, 'last_failover': UNKWN,
                               'primary': UNKWN, 'environment': env.id, "state": "up"},
                           {'name': 'agent4', 'paused': False, 'last_failover': '', 'primary': '',
                            'environment': env2.id, "state": "down"}]}
    assert_equal_ish(shouldbe, all_agents, sortby=["name"])

    code, all_agents = await am.list_agents(env2)
    assert code == 200
    shouldbe = {
        'agents': [{'name': 'agent4', 'paused': False, 'last_failover': '', 'primary': '',
                    'environment': env2.id, "state": "down"}]}
    assert_equal_ish(shouldbe, all_agents)


@pytest.mark.asyncio(timeout=30)
async def test_db_clean(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="testenv", project=project.id)
    await env.insert()
    await data.Agent(environment=env.id, name="agent1", paused=True).insert()
    await data.Agent(environment=env.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env.id, name="agent3", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_future.side_effect = futures
    am = AgentManager(server, False)
    am.add_future = futures
    am.running = True

    async def assert_agent(name: str, state: str, sid: UUID):
        agent = await data.Agent.get(env.id, name)
        assert agent.get_status() == state
        if state == "paused":
            assert agent.primary is None
            assert agent.paused
        elif state == "down":
            assert agent.primary is None
            assert not agent.paused
        elif state == "up":
            assert agent.primary is not None
            agent_instance = await data.AgentInstance.get_by_id(agent.primary)
            agent_proc = await data.AgentProcess.get_one(sid=agent_instance.process)
            assert agent_proc.sid == sid
            assert agent.get_status() == "up"

    async def assert_agents(s1, s2, s3, sid1=None, sid2=None, sid3=None):
        await assert_agent("agent1", s1, sid1)
        await assert_agent("agent2", s2, sid2)
        await assert_agent("agent3", s3, sid3)

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    await futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", True)
    ts1.get_client().reset_mock()
    await assert_agents("paused", "up", "down", sid2=ts1.id)

    # alive
    am.seen(ts1, ["agent1", "agent2"])
    await futures.proccess()
    assert len(am.sessions) == 1
    await assert_agents("paused", "up", "down", sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    await futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", True)
    ts2.get_client().reset_mock()
    await assert_agents("paused", "up", "up", sid2=ts1.id, sid3=ts2.id)

    # expire first
    am.expire(ts1, 100)
    await futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", True)
    ts2.get_client().reset_mock()
    await assert_agents("paused", "up", "up", sid2=ts2.id, sid3=ts2.id)

    # failover
    am = AgentManager(server, False)
    am.add_future = futures
    am.running = True
    await am.clean_db()

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    await futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", True)
    ts1.get_client().reset_mock()
    await assert_agents("paused", "up", "down", sid2=ts1.id)

    # alive
    am.seen(ts1, ["agent1", "agent2"])
    await futures.proccess()
    assert len(am.sessions) == 1
    await assert_agents("paused", "up", "down", sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    await futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", True)
    ts2.get_client().reset_mock()
    await assert_agents("paused", "up", "up", sid2=ts1.id, sid3=ts2.id)

    # expire first
    am.expire(ts1, 100)
    await futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", True)
    ts2.get_client().reset_mock()
    await assert_agents("paused", "up", "up", sid2=ts2.id, sid3=ts2.id)

    # expire second
    am.expire(ts2, 100)
    await futures.proccess()
    assert len(am.sessions) == 0
    await assert_agents("paused", "down", "down")



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from inmanta import resources
import pytest
from inmanta.resources import resource, ResourceException


class Base(resources.Resource):
    fields = ("a", "b", "c")

    @staticmethod
    def get_a(exporter, resource):
        return resource.a


@resources.resource("test_resource::Resource", agent="a", id_attribute="b")
class Resource(Base):
    fields = ("c", "d")
    map = {"d": lambda _, x: x.d}


def test_field_merge():
    assert(len(Resource.fields) == 5)


def test_fields_type():
    with pytest.raises(Exception):
        class Test(resources.Resource):
            fields = ("z")


def test_fields_parent_type():
    with pytest.raises(Exception):
        class Base(resources.Resource):
            fields = ("y")

        class Test(Base):
            fields = ("z",)


def test_resource_base(snippetcompiler):

    import inmanta.resources

    @resource("__config__::XResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "agent")

    snippetcompiler.setup_for_snippet("""
        entity XResource:
            string key
            string agent
            string value
        end

        implement XResource using none

        implementation none for XResource:
        end

        XResource(key="key", agent="agent", value="value")
        """, autostd=False)
    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 1
    myresource = next(json_value.values().__iter__())

    assert myresource.key == "key"
    assert myresource.agent == "agent"
    assert myresource.value == "value"


def test_resource_base_with_method_key(snippetcompiler):

    import inmanta.resources

    @resource("__config__::XResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "agent", "serialize")

        @staticmethod
        def get_serialize(_exporter, resource):
            return resource.key

    snippetcompiler.setup_for_snippet("""
        entity XResource:
            string key
            string agent
            string value
        end

        implement XResource using none

        implementation none for XResource:
        end

        XResource(key="key", agent="agent", value="value")
        """, autostd=False)
    with pytest.raises(ResourceException):
        snippetcompiler.do_export()


def test_resource_with_keyword(snippetcompiler):

    import inmanta.resources

    @resource("__config__::YResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "agent", "model")

        @staticmethod
        def get_model(_exporter, resource):
            return resource.key

    snippetcompiler.setup_for_snippet("""
         entity YResource:
             string key
             string agent
             string value
         end

         implement YResource using none

         implementation none for YResource:
         end

         YResource(key="key", agent="agent", value="value")
         """, autostd=False)

    with pytest.raises(ResourceException):
        snippetcompiler.do_export()


def test_resource_with_private_method(snippetcompiler):

    import inmanta.resources

    @resource("__config__::YResource", agent="agent", id_attribute="key")
    class MyResource(inmanta.resources.Resource):
        """
            A file on a filesystem
        """
        fields = ("__setattr__", "key", "value", "agent")

    snippetcompiler.setup_for_snippet("""
        entity YResource:
            string key
            string agent
            string value
        end

        implement YResource using none

        implementation none for YResource:
        end

        YResource(key="key", agent="agent", value="value")
        """, autostd=False)

    with pytest.raises(ResourceException):
        snippetcompiler.do_export()



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import logging

import pytest

from inmanta import util
from inmanta.util import ensure_future_and_handle_exception
from utils import LogSequence, log_contains, no_error_in_logs

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_scheduler_remove(caplog):
    sched = util.Scheduler("remove")

    i = []

    async def action():
        i.append(0)

    sched.add_action(action, 0.05, 0)

    while len(i) == 0:
        await asyncio.sleep(0.01)

    sched.remove(action)
    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length
    no_error_in_logs(caplog)


@pytest.mark.asyncio
async def test_scheduler_stop(caplog):
    sched = util.Scheduler("stop")

    i = []

    async def action():
        i.append(0)
        return "A"

    sched.add_action(action, 0.05, 0)

    while len(i) == 0:
        await asyncio.sleep(0.01)

    sched.stop()

    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length
    no_error_in_logs(caplog)


@pytest.mark.asyncio
async def test_scheduler_async_run_fail(caplog):
    sched = util.Scheduler("xxx")

    i = []

    async def action():
        i.append(0)
        await asyncio.sleep(0)
        raise Exception("Marker")

    sched.add_action(action, 0.05, 0)

    while len(i) == 0:
        await asyncio.sleep(0.01)

    sched.stop()

    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length

    print(caplog.messages)

    log_contains(
        caplog,
        "inmanta.util",
        logging.ERROR,
        "Uncaught exception while executing scheduled action",
    )


@pytest.mark.asyncio
async def test_scheduler_run_async(caplog):
    sched = util.Scheduler("xxx")

    i = []

    async def action():
        i.append(0)

    sched.add_action(action, 0.05, 0)

    while len(i) == 0:
        await asyncio.sleep(0.01)

    sched.stop()

    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length
    no_error_in_logs(caplog)


@pytest.mark.asyncio
async def test_ensure_future_and_handle_exception(caplog):
    caplog.set_level(logging.INFO)

    async def success():
        LOGGER.info("Success")

    async def fail():
        LOGGER.info("Fail")
        raise Exception("message F")

    ensure_future_and_handle_exception(LOGGER, "marker 1", success())
    ensure_future_and_handle_exception(LOGGER, "marker 2", fail())

    await asyncio.sleep(0.2)

    LogSequence(caplog).log_contains("test_util", logging.INFO, "Success")
    final = (
        LogSequence(caplog)
        .log_contains("test_util", logging.INFO, "Fail")
        .log_contains("test_util", logging.ERROR, "marker 2")
        .index
    )
    exception = caplog.get_records("call")[final].exc_info[1]
    assert str(exception) == "message F"



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest


@pytest.mark.asyncio
async def test_project(server, client, cli):
    # create a new project
    result = await cli.run("project", "create", "-n", "test_project")
    assert result.exit_code == 0

    projects = await client.list_projects()
    assert len(projects.result["projects"]) == 1

    project = projects.result["projects"][0]
    assert project["id"] in result.output
    assert project["name"] in result.output

    # show the project
    result = await cli.run("project", "show", project["id"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert project["name"] in result.output

    result = await cli.run("project", "show", project["name"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert project["name"] in result.output

    # modify the project
    new_name = "test_project_2"
    result = await cli.run("project", "modify", "-n", new_name, project["name"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert new_name in result.output

    new_name = "test_project_3"
    result = await cli.run("project", "modify", "-n", new_name, project["id"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert new_name in result.output

    # delete the project
    result = await cli.run("project", "delete", project["id"])
    assert result.exit_code == 0


@pytest.mark.asyncio
async def test_environment(server, client, cli):
    project_name = "test_project"
    result = await client.create_project(project_name)
    assert result.code == 200
    project_id = result.result["project"]["id"]

    # create a new environment
    result = await cli.run("environment", "create", "-n", "test1", "-r", "/git/repo", "-b", "dev1", "-p", project_name)
    assert result.exit_code == 0

    result = await cli.run("environment", "create", "-n", "test2", "-r", "/git/repo", "-b", "dev2", "-p", project_id)
    assert result.exit_code == 0

    environments = await client.list_environments()
    assert len(environments.result["environments"]) == 2
    environments = environments.result["environments"]

    # list environments
    result = await cli.run("environment", "list")
    assert result.exit_code == 0
    assert "test_project" in result.output
    assert "test1" in result.output
    assert "test2" in result.output

    # show an environment
    env_name = environments[0]["name"]
    env_id = environments[0]["id"]

    result = await cli.run("environment", "show", env_name)
    assert result.exit_code == 0
    assert env_name in result.output
    assert env_id in result.output

    result = await cli.run("environment", "show", env_id)
    assert result.exit_code == 0
    assert env_name in result.output
    assert env_id in result.output


@pytest.mark.asyncio
async def test_environment_settings(server, environment, client, cli):
    result = await cli.run("environment", "setting", "list", "-e", environment)
    assert result.exit_code == 0

    result = await cli.run("environment", "setting", "set", "-e", environment, "-k", "auto_deploy", "-o", "true")
    assert result.exit_code == 0
    result = await cli.run("environment", "setting", "set", "-e", environment, "--key", "auto_deploy", "--value", "true")
    assert result.exit_code == 0

    result = await cli.run("environment", "setting", "list", "-e", environment)
    assert result.exit_code == 0
    assert environment in result.output
    assert "auto_deploy" in result.output

    result = await cli.run("environment", "setting", "get", "-e", environment, "--key", "auto_deploy")
    assert result.exit_code == 0
    assert "True" in result.output

    result = await cli.run("environment", "setting", "delete", "-e", environment, "--key", "auto_deploy")
    assert result.exit_code == 0


@pytest.mark.asyncio
async def test_agent(server, client, environment, cli):
    result = await cli.run("agent", "list", "-e", environment)
    assert result.exit_code == 0


@pytest.mark.parametrize("push_method", [([]),
                                         (["-p"]),
                                         (["-p", "--full"])])
@pytest.mark.asyncio
async def test_version(server, client, environment, cli, push_method):
    version = "12345"
    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=' + version,
                  'send_event': False,
                  'purged': False,
                  'requires': ['test::Resource[agent1,key=key2],v=' + version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent1,key=key2],v=' + version,
                  'send_event': False,
                  'requires': [],
                  'purged': False,
                  },
                 {'key': 'key3',
                  'value': None,
                  'id': 'test::Resource[agent1,key=key3],v=' + version,
                  'send_event': False,
                  'requires': [],
                  'purged': True,
                  }
                 ]

    result = await client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    result = await cli.run("version", "list", "-e", environment)
    assert result.exit_code == 0
    assert version in result.output
    assert "pending" in result.output

    result = await cli.run("version", "release", "-e", environment, version, *push_method)
    assert result.exit_code == 0
    assert version in result.output

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["result"] == "deploying"

    result = await cli.run("version", "report", "-e", environment, "-i", version)
    assert result.exit_code == 0


@pytest.mark.asyncio
async def test_param(server, client, environment, cli):
    result = await cli.run("param", "set", "-e", environment, "--name", "var1", "--value", "value1")
    assert result.exit_code == 0
    assert "value1" in result.output

    result = await cli.run("param", "get", "-e", environment, "--name", "var1")
    assert result.exit_code == 0
    assert "value1" in result.output

    result = await cli.run("param", "list", "-e", environment)
    assert result.exit_code == 0
    assert "var1" in result.output


@pytest.mark.asyncio
async def test_form_and_records(server, client, environment, cli):
    form_type = "FormType"
    result = await client.put_form(tid=environment, id=form_type,
                                   form={'attributes': {'field1': {'default': 1, 'options': {'min': 1, 'max': 100},
                                                                   'type': 'number'},
                                                        'field2': {'default': "", 'options': {}, 'type': 'string'}},
                                         'options': {},
                                         'type': form_type}
                                   )
    assert(result.code == 200)
    form_id = result.result["form"]["id"]

    result = await cli.run("form", "list", "-e", environment)
    assert result.exit_code == 0
    assert form_id in result.output

    result = await cli.run("form", "show", "-e", environment, "-t", form_type)
    assert result.exit_code == 0
    assert "field1" in result.output
    assert "field2" in result.output

    result = await cli.run("record", "create", "-e", environment, "-t", form_type, "-p", "field1=1234", "-p", "field2=test456")
    assert result.exit_code == 0
    assert "1234" in result.output
    assert "test456" in result.output

    records = await client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    record = records.result["records"][0]
    record_id = record["id"]

    result = await cli.run("record", "list", "-e", environment, "-t", form_type)
    assert result.exit_code == 0
    assert record_id in result.output

    result = await cli.run("record", "list", "-e", environment, "-t", form_type, "--show-all")
    assert result.exit_code == 0

    result = await cli.run("record", "update", "-e", environment, "-r", record_id, "-p", "field1=98765")
    assert result.exit_code == 0
    assert "98765" in result.output

    result = await cli.run("record", "delete", "-e", environment, record_id)
    assert result.exit_code == 0

    records = await client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    assert len(records.result["records"]) == 0


@pytest.mark.asyncio
async def test_import_export(server, client, environment, cli, tmpdir):
    form_type = "FormType"
    result = await client.put_form(tid=environment, id=form_type,
                                   form={'attributes': {'field1': {'default': 1, 'options': {'min': 1, 'max': 100},
                                                                   'type': 'number'},
                                                        'field2': {'default': "", 'options': {}, 'type': 'string'}},
                                         'options': {},
                                         'type': form_type}
                                   )
    form_id = result.result["form"]["id"]

    result = await cli.run("record", "create", "-e", environment, "-t", form_type, "-p", "field1=1234", "-p", "field2=test456")
    assert result.exit_code == 0

    result = await cli.run("form", "export", "-e", environment, "-t", form_type)
    assert form_id in result.output

    f = tmpdir.join("export.json")
    f.write(result.output)

    records = await client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    record = records.result["records"][0]

    await client.delete_record(tid=environment, id=record["id"])
    records = await client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    assert len(records.result["records"]) == 0

    result = await cli.run("form", "import", "-e", environment, "-t", form_type, "--file", str(f))

    records = await client.list_records(tid=environment, form_type=form_type)
    assert records.code == 200
    assert len(records.result["records"]) == 1



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: bart@inmanta.com
"""

from subprocess import CalledProcessError


from inmanta import env
import pytest


def test_basic_install(tmpdir):

    env_dir1 = tmpdir.mkdir("env1").strpath

    with pytest.raises(ImportError):
        import lorem  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    venv1._install(["lorem"])
    import lorem  # NOQA
    lorem.sentence()

    with pytest.raises(ImportError):
        import yummy  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    venv1.install_from_list(["dummy-yummy"])
    import yummy  # NOQA

    with pytest.raises(ImportError):
        import iplib  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    try:
        venv1.install_from_list(
            ["lorem == 0.1.1", "dummy-yummy", "iplib@git+https://github.com/bartv/python3-iplib", "lorem", "iplib >=0.0.1"])
    except CalledProcessError as ep:
        print(ep.stdout)
        raise
    import iplib  # NOQA


def test_req_parser(tmpdir):
    url = "git+https://github.com/bartv/python3-iplib"
    at_url = "iplib@" + url
    egg_url = url + "#egg=iplib"

    e = env.VirtualEnv(tmpdir)
    name, u = e._parse_line(url)
    assert(name is None)
    assert(u == url)

    name, u = e._parse_line(at_url)
    assert(name == "iplib")
    assert(u == egg_url)

    e._parse_line(egg_url)
    assert(name == "iplib")
    assert(u == egg_url)


def test_gen_req_file(tmpdir):
    e = env.VirtualEnv(tmpdir)
    req = ["lorem == 0.1.1", "lorem > 0.1", "dummy-yummy", "iplib@git+https://github.com/bartv/python3-iplib", "lorem"]

    req_lines = [x for x in e._gen_requirements_file(req).split("\n") if len(x) > 0]
    assert(len(req_lines) == 3)
